package cn.com.qmth.examcloud.core.oe.admin.api.controller;

import cn.com.qmth.examcloud.commons.helpers.poi.ExcelWriter;
import cn.com.qmth.examcloud.core.oe.admin.service.ExamStatisticService;
import cn.com.qmth.examcloud.core.oe.admin.service.bean.statistic.ExamStatisticInfo;
import cn.com.qmth.examcloud.web.config.SystemProperties;
import cn.com.qmth.examcloud.web.support.ControllerSupport;
import cn.com.qmth.examcloud.web.support.Naked;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import org.apache.commons.io.FileUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.*;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

@RestController
@Api(tags = "考试统计相关接口")
@RequestMapping("${$rmp.ctr.oe}/exam/statistic")
public class ExamStatisticController extends ControllerSupport {

    @Autowired
    private ExamStatisticService examStatisticService;

    @Autowired
    private SystemProperties systemConfig;

    @PostMapping("/overview/refresh")
    @ApiOperation(value = "统计概况（刷新）")
    public void refresh(@RequestParam Long examId) {
        examStatisticService.refresh(examId);
    }

    @PostMapping("/overview/for/exam")
    @ApiOperation(value = "统计概况（考试）")
    public ExamStatisticInfo overviewForExam(@RequestParam Long examId) {
        return examStatisticService.overviewForExam(examId);
    }

    @PostMapping("/overview/for/course")
    @ApiOperation(value = "统计概况（课程）")
    public List<ExamStatisticInfo> overviewForCourse(@RequestParam Long examId, @RequestParam(required = false) Long courseId) {
        return examStatisticService.overviewForCourse(examId, courseId);
    }

    @PostMapping("/overview/for/org")
    @ApiOperation(value = "统计概况（学习中心）")
    public List<ExamStatisticInfo> overviewForOrg(@RequestParam Long examId, @RequestParam(required = false) Long orgId) {
        return examStatisticService.overviewForOrg(examId, orgId);
    }

    @Naked
    @GetMapping("/overview/for/course/export")
    @ApiOperation(value = "统计概况（课程）导出Excel")
    public void overviewForCourseExport(@RequestParam Long examId, @RequestParam(required = false) Long courseId) throws Exception {
        ExamStatisticInfo examOverview = examStatisticService.overviewForExam(examId);

        List<ExamStatisticInfo> list = new ArrayList<>();
        if (examOverview.getExamId() != null) {
            list.add(examOverview);
            list.addAll(examStatisticService.overviewForCourse(examId, courseId));
        }

        List<Object[]> lines = new ArrayList<>();
        for (ExamStatisticInfo info : list) {
            lines.add(new Object[]{info.getExamId(), info.getExamName(), info.getCourseName(), info.getCourseCode(),
                    info.getAllCount(), info.getFinishCount(), info.getUnFinishCount(), info.getUnFinishRate(),
                    info.getPassScoreCount(), info.getPassScoreRate(), info.getGoodScoreCount(), info.getGoodScoreRate(),
                    info.getIllegalityCount(), info.getIllegalityRate()});
        }

        String filePath = systemConfig.getTempDataDir() + File.separator + System.currentTimeMillis() + ".xlsx";
        File file = new File(filePath);

        ExcelWriter.write(new String[]{"考试ID", "考试名称", "课程名称", "课程代码", "应考人数", "实考人数", "缺考人数", "缺考率（%）",
                        "及格人数", "及格率（%）", "优秀人数", "优秀率（%）", "违纪人数", "违纪率（%）"},
                new Class[]{Long.class, String.class, String.class, String.class, Integer.class, Integer.class,
                        Integer.class, Double.class, Integer.class, Double.class, Integer.class, Double.class,
                        Integer.class, Double.class},
                lines, new File(filePath));

        exportFile("课程统计.xlsx", file);
        FileUtils.deleteQuietly(file);
    }

    @Naked
    @GetMapping("/overview/for/org/export")
    @ApiOperation(value = "统计概况（学习中心）导出Excel")
    public void overviewForOrgExport(@RequestParam Long examId, @RequestParam(required = false) Long orgId) throws Exception {
        ExamStatisticInfo examOverview = examStatisticService.overviewForExam(examId);

        List<ExamStatisticInfo> list = new ArrayList<>();
        if (examOverview.getExamId() != null) {
            list.add(examOverview);
            list.addAll(examStatisticService.overviewForOrg(examId, orgId));
        }

        List<Object[]> lines = new ArrayList<>();
        for (ExamStatisticInfo info : list) {
            lines.add(new Object[]{info.getExamId(), info.getExamName(), info.getOrgName(),
                    info.getAllCount(), info.getFinishCount(), info.getUnFinishCount(), info.getUnFinishRate(),
                    info.getPassScoreCount(), info.getPassScoreRate(), info.getGoodScoreCount(), info.getGoodScoreRate(),
                    info.getIllegalityCount(), info.getIllegalityRate()});
        }

        String filePath = systemConfig.getTempDataDir() + File.separator + System.currentTimeMillis() + ".xlsx";
        File file = new File(filePath);

        ExcelWriter.write(new String[]{"考试ID", "考试名称", "学习中心", "应考人数", "实考人数", "缺考人数", "缺考率（%）",
                        "及格人数", "及格率（%）", "优秀人数", "优秀率（%）", "违纪人数", "违纪率（%）"},
                new Class[]{Long.class, String.class, String.class, Integer.class, Integer.class, Integer.class,
                        Double.class, Integer.class, Double.class, Integer.class, Double.class,
                        Integer.class, Double.class},
                lines, new File(filePath));

        exportFile("学习中心统计.xlsx", file);
        FileUtils.deleteQuietly(file);
    }

}
