package cn.com.qmth.examcloud.core.oe.admin.service.impl;

import cn.com.qmth.examcloud.api.commons.enums.ExamType;
import cn.com.qmth.examcloud.core.oe.admin.dao.ExamRecordDataRepo;
import cn.com.qmth.examcloud.core.oe.admin.dao.ExamRecordForMarkingRepo;
import cn.com.qmth.examcloud.core.oe.admin.dao.entity.ExamRecordDataEntity;
import cn.com.qmth.examcloud.core.oe.admin.dao.entity.ExamRecordForMarkingEntity;
import cn.com.qmth.examcloud.core.oe.admin.dao.enums.MarkingType;
import cn.com.qmth.examcloud.core.oe.admin.service.ExamRecordForMarkingService;
import cn.com.qmth.examcloud.core.oe.admin.service.ExamRecordQuestionsService;
import cn.com.qmth.examcloud.core.oe.admin.service.bean.examrecord.ExamRecordDataBean;
import cn.com.qmth.examcloud.support.enums.ExamProperties;
import cn.com.qmth.examcloud.support.helper.ExamCacheTransferHelper;
import cn.com.qmth.examcloud.web.helpers.GlobalHelper;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.stereotype.Service;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.*;
import java.util.Map.Entry;
import java.util.stream.Collectors;

/**
 * @author chenken
 * @date 2018年9月19日 下午3:20:29
 * @company QMTH
 * @description ExamRecordForMarkingServiceImpl.java
 */
@Service("examRecordForMarkingService")
public class ExamRecordForMarkingServiceImpl implements ExamRecordForMarkingService {

    @Autowired
    private ExamRecordForMarkingRepo examRecordForMarkingRepo;

    @Autowired
    private ExamRecordDataRepo examRecordDataRepo;

    @Autowired
    private ExamRecordQuestionsService examRecordQuestionsService;

    @Autowired
    private JdbcTemplate jdbcTemplate;

    @Override
    public List<ExamRecordForMarkingEntity> queryValidExamRecordInfoByStuIds(
            Long examId, Long courseId, List<Long> examStudentIds, String batchNum) {
        ExamType examType = getExamType(examId);

        List<ExamRecordForMarkingEntity> examRecordForMarkingList;

        String markingType = getMarkingType(examId);
        if (StringUtils.isEmpty(batchNum)) {
            examRecordForMarkingList =
                    examRecordForMarkingRepo.findByExamIdAndCourseIdAndExamStudentIds(examId, courseId, examStudentIds);
        } else {
            if ((ExamType.ONLINE == examType || ExamType.ONLINE_HOMEWORK == examType) &&
                    (MarkingType.OBJECT_SCORE_MAX.name().equals(markingType) ||
                            MarkingType.LAST_SUBMIT.name().equals(markingType))) {
                examRecordForMarkingList =
                        examRecordForMarkingRepo.findByExamIdAndCourseIdAndExamStudentIds(examId, courseId, examStudentIds);
            } else {
                examRecordForMarkingList =
                        examRecordForMarkingRepo.findByExamIdAndCourseIdAndExamStudentIdsAndBatchNum(examId, courseId, examStudentIds, batchNum);
            }
        }
        if (examType == ExamType.OFFLINE) {
            return examRecordForMarkingList;
        } else if (ExamType.ONLINE == examType || ExamType.ONLINE_HOMEWORK == examType) {
            if (markingType.equals(MarkingType.ALL.name())) {
                return examRecordForMarkingList;
            } else if (markingType.equals(MarkingType.OBJECT_SCORE_MAX.name())) {
                return queryStudentPapersByObjectScoreMax(examRecordForMarkingList);
            } else if (markingType.equals(MarkingType.LAST_SUBMIT.name())) {
                return queryStudentPapersByLastSubmit(examRecordForMarkingList);
            }
        }
        return null;
    }

    /**
     * 获取阅卷类型
     *
     * @param examId
     * @return
     */
    private String getMarkingType(Long examId) {
        return ExamCacheTransferHelper.
                getDefaultCachedExamProperty(examId, ExamProperties.MARKING_TYPE.name()).getValue();
    }

    @Override
    public List<ExamRecordForMarkingEntity> queryValidExamRecordList(Long examId, Long courseId) {
        ExamType examType = getExamType(examId);

        List<ExamRecordForMarkingEntity> examRecordForMarkingList = examRecordForMarkingRepo.findByExamIdAndCourseId(examId, courseId);
        if (examType == ExamType.OFFLINE) {
            return examRecordForMarkingList;
        } else if (ExamType.ONLINE == examType || ExamType.ONLINE_HOMEWORK == examType) {
            String markingType = getMarkingType(examId);

            if (markingType.equals(MarkingType.ALL.name())) {
                return examRecordForMarkingList;
            } else if (markingType.equals(MarkingType.OBJECT_SCORE_MAX.name())) {
                return queryStudentPapersByObjectScoreMax(examRecordForMarkingList);
            } else if (markingType.equals(MarkingType.LAST_SUBMIT.name())) {
                return queryStudentPapersByLastSubmit(examRecordForMarkingList);
            }
        }
        return null;
    }

    /**
     * 获取考试类型
     *
     * @param examId
     * @return
     */
    private ExamType getExamType(Long examId) {
        return ExamType.valueOf(ExamCacheTransferHelper.getDefaultCachedExam(examId).getExamType());
    }

    /**
     * 最后一次提交
     *
     * @param examRecordForMarkingAllList
     * @return
     */
    private List<ExamRecordForMarkingEntity> queryStudentPapersByLastSubmit(List<ExamRecordForMarkingEntity> examRecordForMarkingAllList) {
        Map<Long, List<ExamRecordForMarkingEntity>> groupByExamStudentIdMap = groupByExamStudentId(examRecordForMarkingAllList);
        Iterator<Entry<Long, List<ExamRecordForMarkingEntity>>> iterator = groupByExamStudentIdMap.entrySet().iterator();
        List<ExamRecordForMarkingEntity> finalExamRecordForMarkingEntityList = new ArrayList<ExamRecordForMarkingEntity>();
        while (iterator.hasNext()) {
            Entry<Long, List<ExamRecordForMarkingEntity>> entry = iterator.next();
            List<ExamRecordForMarkingEntity> examRecordForMarkingList = entry.getValue();
            List<ExamRecordForMarkingEntity> listSortByExamRecordDataId = examRecordForMarkingList.stream().sorted((p1, p2) -> p2.getExamRecordDataId().compareTo(p1.getExamRecordDataId())).collect(Collectors.toList());
            finalExamRecordForMarkingEntityList.add(listSortByExamRecordDataId.get(0));
        }
        /**
         * 解决下面的问题:
         *  当一个课程下设置两套试卷一套有主观题，一套没有主观题时，考试完成后给阅卷端传试卷时会传错，例如：
         *	考试批次设置为取最后一次考试成绩；
         *	在调卷规则中给课程A设置了两份试卷A1和A2，A1试卷中有主观题，A2试卷没有主观题；
         *	考生AA在考试时，第一次抽取到了A1试卷，第二次抽取到了A2试卷；
         *	考试完成后考生AA应该没有试卷传给阅卷端，但实际上系统将A1试卷传给了阅卷端
         */
        //todo
        return finalExamRecordForMarkingEntityList;
    }

    /**
     * 客观分最高
     *
     * @param examRecordForMarkingAllList
     * @return
     */
    private List<ExamRecordForMarkingEntity> queryStudentPapersByObjectScoreMax(List<ExamRecordForMarkingEntity> examRecordForMarkingAllList) {
        Map<Long, List<ExamRecordForMarkingEntity>> groupByExamStudentIdMap = groupByExamStudentId(examRecordForMarkingAllList);
        Iterator<Entry<Long, List<ExamRecordForMarkingEntity>>> iterator = groupByExamStudentIdMap.entrySet().iterator();
        List<ExamRecordForMarkingEntity> finalExamRecordForMarkingEntityList = new ArrayList<ExamRecordForMarkingEntity>();
        while (iterator.hasNext()) {
            Entry<Long, List<ExamRecordForMarkingEntity>> entry = iterator.next();
            List<ExamRecordForMarkingEntity> examRecordForMarkingList = entry.getValue();
            //得出该考生的最高客观分
            double maxObjectiveScore = examRecordForMarkingList.stream().sorted((p1, p2) -> p2.getObjectiveScore().compareTo(p1.getObjectiveScore()))
                    .limit(1).findFirst().get().getObjectiveScore();
            //过滤出最高分集合
            List<ExamRecordForMarkingEntity> examRecordForMarkingByFilterList = examRecordForMarkingList.stream().filter(score -> {
                return score.getObjectiveScore() == maxObjectiveScore;
            }).collect(Collectors.toList());
            //如果最高客观分的有多个,继续按主观题答案长度过滤
            if (examRecordForMarkingByFilterList.size() > 1) {
                int maxAnswerLength = examRecordForMarkingByFilterList.stream().sorted((p1, p2) -> p2.getSubjectiveAnswerLength() - p1.getSubjectiveAnswerLength())
                        .limit(1).findFirst().get().getSubjectiveAnswerLength();
                examRecordForMarkingByFilterList = examRecordForMarkingList.stream().filter(score -> {
                    return score.getSubjectiveAnswerLength() == maxAnswerLength;
                }).collect(Collectors.toList());
                //如果主观题答案最长长度一样的还有多个，继续按时间(考试记录ID)过滤出最新的那条
                if (examRecordForMarkingByFilterList.size() > 1) {
                    examRecordForMarkingByFilterList = examRecordForMarkingByFilterList.stream().sorted((p1, p2) -> p2.getExamRecordDataId().compareTo(p1.getExamRecordDataId()))
                            .limit(1)
                            .collect(Collectors.toList());
                }
            }
            finalExamRecordForMarkingEntityList.add(examRecordForMarkingByFilterList.get(0));
        }
        return finalExamRecordForMarkingEntityList;
    }

    private Map<Long, List<ExamRecordForMarkingEntity>> groupByExamStudentId(List<ExamRecordForMarkingEntity> examRecordForMarkingList) {
        Map<Long, List<ExamRecordForMarkingEntity>> groupByExamStudentIdMap = new HashMap<Long, List<ExamRecordForMarkingEntity>>();
        for (ExamRecordForMarkingEntity examRecordForMarkingEntity : examRecordForMarkingList) {
            Long examStudentId = examRecordForMarkingEntity.getExamStudentId();
            List<ExamRecordForMarkingEntity> value = null;
            if (groupByExamStudentIdMap.containsKey(examStudentId)) {
                value = groupByExamStudentIdMap.get(examStudentId);
            } else {
                value = new ArrayList<ExamRecordForMarkingEntity>();
            }
            value.add(examRecordForMarkingEntity);
            groupByExamStudentIdMap.put(examStudentId, value);
        }
        return groupByExamStudentIdMap;
    }

    @Override
    public List<ExamRecordForMarkingEntity> findExamRecordForMarkingInfo(Long id, Long examId, Long courseId, String batchNum) {
        if (id == null && examId == null) {
            return null;
        }
        StringBuffer sql = new StringBuffer();
        sql.append("select t.id,t.exam_id,t.exam_record_data_id,t.base_paper_id,t.paper_type,t.course_id,t.batch_num " +
                " from ec_oe_exam_record_4_marking t left join ec_oe_exam_record_data f  "
        		+" on t.exam_record_data_id=f.id  where f.random_paper=0 ");
        if (id != null) {
            sql.append(" and t.id = " + id);
        }
        if (examId != null) {
            sql.append(" and t.exam_id = " + examId);
        }
        if (courseId != null) {
            sql.append(" and t.course_id = '" + courseId + "'");
        }
        if (StringUtils.isNotEmpty(batchNum)) {
            sql.append(" and (t.batch_num is null or t.batch_num != '" + batchNum + "')");
        }
        sql.append(" group by t.base_paper_id");
        return jdbcTemplate.query(sql.toString(), new RowMapper<ExamRecordForMarkingEntity>() {
            @Override
            public ExamRecordForMarkingEntity mapRow(ResultSet rs, int arg1) throws SQLException {
                ExamRecordForMarkingEntity entity = new ExamRecordForMarkingEntity();
                entity.setId(rs.getLong("id"));
                entity.setExamId(rs.getLong("exam_id"));
                entity.setBasePaperId(rs.getString("base_paper_id"));
                entity.setPaperType(rs.getString("paper_type"));
                entity.setCourseId(rs.getLong("course_id"));
                entity.setBatchNum(rs.getString("batch_num"));
                entity.setExamRecordDataId(rs.getLong("exam_record_data_id"));
                return entity;
            }
        });
    }

    @Override
    public void saveOffLineExamRecordForMarking(ExamRecordDataBean examRecordData) {
        ExamRecordForMarkingEntity examRecordForMarking = examRecordForMarkingRepo.findByExamRecordDataId(examRecordData.getId());
        if (examRecordForMarking == null) {
            examRecordForMarking = new ExamRecordForMarkingEntity();
            examRecordForMarking.setCreationTime(new Date());
        }

        examRecordForMarking.setExamId(examRecordData.getExamId());
        examRecordForMarking.setExamRecordDataId(examRecordData.getId());
        examRecordForMarking.setExamStudentId(examRecordData.getExamStudentId());
        examRecordForMarking.setCourseId(examRecordData.getCourseId());
        examRecordForMarking.setObjectiveScore(0D);
        examRecordForMarking.setSubjectiveAnswerLength(0);
        examRecordForMarking.setBasePaperId(examRecordData.getBasePaperId());
        examRecordForMarking.setPaperType(examRecordData.getPaperType());
        examRecordForMarking.setUpdateTime(new Date());
        examRecordForMarkingRepo.save(examRecordForMarking);
    }

    /**
     * 在线考试-保存阅卷相关数据-对内
     *
     * @param examRecordDataId
     */
    @Override
    public void saveExamRecordForMarking(Long examRecordDataId, Double objectiveScore) {
        ExamRecordDataEntity examRecordData =
                GlobalHelper.getEntity(examRecordDataRepo, examRecordDataId, ExamRecordDataEntity.class);

        //全客观题卷
        if (null == examRecordData.getIsAllObjectivePaper() ||
                (null != examRecordData.getIsAllObjectivePaper() && examRecordData.getIsAllObjectivePaper())) {
            return;
        }
        //违纪
        if (null != examRecordData.getIsIllegality() && examRecordData.getIsIllegality()) {
            return;
        }
        //有警告未审核
        if (null != examRecordData.getIsWarn() && examRecordData.getIsWarn()
                && null != examRecordData.getIsAudit() && !examRecordData.getIsAudit()) {
            return;
        }
        //已经存在
        ExamRecordForMarkingEntity examRecordForMarkingExists = examRecordForMarkingRepo.findByExamRecordDataId(examRecordData.getId());
        if (examRecordForMarkingExists != null) {
            return;
        }

        ExamRecordForMarkingEntity examRecordForMarking = new ExamRecordForMarkingEntity();
        examRecordForMarking.setExamId(examRecordData.getExamId());
        examRecordForMarking.setExamRecordDataId(examRecordData.getId());
        examRecordForMarking.setExamStudentId(examRecordData.getExamStudentId());
        examRecordForMarking.setBasePaperId(examRecordData.getBasePaperId());
        examRecordForMarking.setPaperType(examRecordData.getPaperType());
        examRecordForMarking.setCourseId(examRecordData.getCourseId());
        examRecordForMarking.setObjectiveScore(objectiveScore);

        int subjectiveAnswerLength = examRecordQuestionsService.calculationSubjectiveAnswerLength(examRecordData.getId());
        examRecordForMarking.setSubjectiveAnswerLength(subjectiveAnswerLength);

        examRecordForMarkingRepo.save(examRecordForMarking);
    }

    /**
     * 获取某个考生的有效考试记录
     *
     * @param examStudentId 考生id集合
     * @return
     */
    @Override
    public List<ExamRecordForMarkingEntity> queryValidExamRecordList(Long examStudentId) {
        List<ExamRecordForMarkingEntity> examRecordForMarkingList =
                examRecordForMarkingRepo.findByExamStudentId(examStudentId);

        if (null == examRecordForMarkingList || examRecordForMarkingList.isEmpty()) {
            return null;
        }

        Long examId = examRecordForMarkingList.get(0).getExamId();
        ExamType examType = getExamType(examId);

        if (examType == ExamType.OFFLINE) {
            return examRecordForMarkingList;
        } else if (ExamType.ONLINE == examType || ExamType.ONLINE_HOMEWORK == examType) {

            String markingType = getMarkingType(examId);
            if (markingType.equals(MarkingType.ALL.name())) {
                return examRecordForMarkingList;
            } else if (markingType.equals(MarkingType.OBJECT_SCORE_MAX.name())) {
                return queryStudentPapersByObjectScoreMax(examRecordForMarkingList);
            } else if (markingType.equals(MarkingType.LAST_SUBMIT.name())) {
                return queryStudentPapersByLastSubmit(examRecordForMarkingList);
            }
        }
        return null;
    }

}
