package cn.com.qmth.examcloud.core.oe.admin.service.impl;

import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.util.MathUtils;
import cn.com.qmth.examcloud.core.basic.api.CourseCloudService;
import cn.com.qmth.examcloud.core.basic.api.OrgCloudService;
import cn.com.qmth.examcloud.core.basic.api.bean.CourseBean;
import cn.com.qmth.examcloud.core.basic.api.bean.OrgBean;
import cn.com.qmth.examcloud.core.basic.api.request.GetCoursesByIdListReq;
import cn.com.qmth.examcloud.core.basic.api.request.GetOrgMapsReq;
import cn.com.qmth.examcloud.core.basic.api.response.GetCoursesByIdListResp;
import cn.com.qmth.examcloud.core.oe.admin.dao.ExamStatisticRepo;
import cn.com.qmth.examcloud.core.oe.admin.dao.entity.ExamStatisticEntity;
import cn.com.qmth.examcloud.core.oe.admin.service.ExamStatisticService;
import cn.com.qmth.examcloud.core.oe.admin.service.bean.statistic.ExamStatisticInfo;
import cn.com.qmth.examcloud.core.oe.admin.service.bean.statistic.ExamStudentScoreInfo;
import cn.com.qmth.examcloud.examwork.api.ExamCloudService;
import cn.com.qmth.examcloud.examwork.api.bean.ExamBean;
import cn.com.qmth.examcloud.examwork.api.request.GetExamCourseIdsReq;
import cn.com.qmth.examcloud.examwork.api.request.GetExamCourseReq;
import cn.com.qmth.examcloud.examwork.api.request.GetExamMapsReq;
import cn.com.qmth.examcloud.examwork.api.response.GetExamCourseIdsResp;
import cn.com.qmth.examcloud.examwork.api.response.GetExamCourseResp;
import cn.com.qmth.examcloud.examwork.api.response.GetExamMapsResp;
import cn.com.qmth.examcloud.support.CacheConstants;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import org.apache.commons.collections4.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.BeanPropertyRowMapper;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Service;

import java.util.*;
import java.util.stream.Collectors;

@Service
public class ExamStatisticServiceImpl implements ExamStatisticService {

    private static final Logger log = LoggerFactory.getLogger(ExamStatisticServiceImpl.class);

    @Autowired
    private ExamStatisticRepo examStatisticRepo;

    @Autowired
    private CourseCloudService courseCloudService;

    @Autowired
    private OrgCloudService orgCloudService;

    @Autowired
    private ExamCloudService examCloudService;

    @Autowired
    private RedisClient redisClient;

    @Autowired
    private JdbcTemplate jdbcTemplate;

    private final static int PASS_SCORE_LINE = 60;// 及格率默认值

    private final static int GOOD_SCORE_LINE = 90;// 优秀率默认值

    @Override
    public void refresh(Long examId) {
        final String cacheKey = CacheConstants.CACHE_OE_SCORE_STATISTIC_STATUS + examId;
        Boolean running = redisClient.get(cacheKey, Boolean.class);
        if (running != null) {
            throw new StatusException("500", "统计计算中，请稍后查询结果！");
        }

        try {
            long start = System.currentTimeMillis();
            redisClient.set(cacheKey, true, 600);// 10分钟后自动过期

            // 获取考试相关的课程ID列表
            GetExamCourseIdsReq idsReq = new GetExamCourseIdsReq();
            idsReq.setExamId(examId);
            GetExamCourseIdsResp idsResp = examCloudService.getExamCourseIds(idsReq);
            List<Long> courseIds = idsResp.getCourseIds();
            if (CollectionUtils.isEmpty(courseIds)) {
                return;
            }

            for (Long courseId : courseIds) {
                long courseStart = System.currentTimeMillis();

                GetExamCourseReq settingReq = new GetExamCourseReq();
                settingReq.setExamId(examId);
                settingReq.setCourseId(courseId);
                GetExamCourseResp settingResp = examCloudService.getExamCourseSetting(settingReq);

                int passScoreLine = PASS_SCORE_LINE, goodScoreLine = GOOD_SCORE_LINE;
                if (settingResp != null && settingResp.getBean() != null) {
                    passScoreLine = settingResp.getBean().getPassScoreLine();
                    goodScoreLine = settingResp.getBean().getGoodScoreLine();
                }

                StringBuffer sql = new StringBuffer();
                sql.append(" select es.exam_id,es.course_id,es.org_id,es.exam_student_id,");
                sql.append(" es.finished,sc.total_score,d.paper_score,d.is_illegality as illegality,");
                sql.append(" d.id as exam_record_data_id");
                sql.append(" from ec_oe_exam_student es");
                sql.append(" left join ec_oe_exam_student_final_score sc on sc.exam_student_id = es.exam_student_id");
                sql.append(" left join ec_oe_exam_record_data d on d.id = sc.exam_record_data_id");
                sql.append(" where es.exam_id = ").append(examId);
                sql.append(" and es.course_id = ").append(courseId);

                List<ExamStudentScoreInfo> list = jdbcTemplate.query(sql.toString(), new BeanPropertyRowMapper(ExamStudentScoreInfo.class));
                if (CollectionUtils.isEmpty(list)) {
                    log.warn("ExamStudent is empty! examId:{} courseId:{}", examId, courseId);
                    continue;
                }

                Map<Long, List<ExamStudentScoreInfo>> orgStatisticMaps = list.stream().collect(Collectors.groupingBy(ExamStudentScoreInfo::getOrgId));
                for (Map.Entry<Long, List<ExamStudentScoreInfo>> e : orgStatisticMaps.entrySet()) {
                    int allCount = 0, finishCount = 0, passScoreCount = 0, goodScoreCount = 0, illegalityCount = 0;
                    for (ExamStudentScoreInfo v : e.getValue()) {
                        allCount++;

                        if (v.getFinished() == null || !v.getFinished()) {
                            // 考生缺考，未完成考试
                            continue;
                        }

                        // 考生非缺考，有考试记录
                        finishCount++;

                        if (v.getTotalScore() == null || v.getExamRecordDataId() == null) {
                            // 无最终成绩，算违纪
                            illegalityCount++;
                        } else {
                            // 有最终成绩，违纪按最终成绩对应的考试记录是否违纪来计算
                            if (v.getIllegality() != null && v.getIllegality()) {
                                illegalityCount++;
                            }

                            if (v.getPaperScore() != null) {
                                if (v.getTotalScore() >= (v.getPaperScore() * passScoreLine) / 100d) {
                                    passScoreCount++;
                                }

                                if (v.getTotalScore() >= (v.getPaperScore() * goodScoreLine) / 100d) {
                                    goodScoreCount++;
                                }
                            }
                        }
                    }

                    ExamStatisticEntity result = examStatisticRepo.findByExamIdAndCourseIdAndOrgId(examId, courseId, e.getKey());
                    if (result == null) {
                        result = new ExamStatisticEntity();
                        result.setExamId(examId);
                        result.setCourseId(courseId);
                        result.setOrgId(e.getKey());
                    }
                    result.setAllCount(allCount);
                    result.setFinishCount(finishCount);
                    result.setPassScoreCount(passScoreCount);
                    result.setGoodScoreCount(goodScoreCount);
                    result.setIllegalityCount(illegalityCount);
                    result.setUpdateTime(new Date());
                    examStatisticRepo.save(result);
                }

                log.warn("refresh ok! examId:{} courseId:{} cost:{}ms", examId, courseId, System.currentTimeMillis() - courseStart);
            }

            log.warn("refresh ok! examId:{} cost:{}ms", examId, System.currentTimeMillis() - start);
        } finally {
            redisClient.delete(cacheKey);
        }
    }

    @Override
    public ExamStatisticInfo overviewForExam(Long examId) {
        StringBuffer sql = new StringBuffer();
        sql.append(" select");
        sql.append(" exam_id,");
        sql.append(" sum(all_count) as allCount,");
        sql.append(" sum(finish_count) as finishCount,");
        sql.append(" sum(pass_score_count) as passScoreCount,");
        sql.append(" sum(good_score_count) as goodScoreCount,");
        sql.append(" sum(illegality_count) as illegalityCount");
        sql.append(" from ec_oe_exam_statistic");
        sql.append(" where exam_id = ").append(examId);

        List<ExamStatisticInfo> list = jdbcTemplate.query(sql.toString(), new BeanPropertyRowMapper(ExamStatisticInfo.class));
        if (list.isEmpty()) {
            return new ExamStatisticInfo();
        }

        ExamStatisticInfo info = list.get(0);
        if (info.getExamId() == null) {
            return new ExamStatisticInfo();
        }

        info.setUnFinishCount(info.getAllCount() - info.getFinishCount());
        if (info.getAllCount() > 0) {
            info.setUnFinishRate(MathUtils.percentage(info.getUnFinishCount(), info.getAllCount()));
            info.setPassScoreRate(MathUtils.percentage(info.getPassScoreCount(), info.getFinishCount()));
            info.setGoodScoreRate(MathUtils.percentage(info.getGoodScoreCount(), info.getFinishCount()));
            info.setIllegalityRate(MathUtils.percentage(info.getIllegalityCount(), info.getFinishCount()));
        }

        // 获取考试信息
        Set<Long> examIds = new HashSet<>();
        examIds.add(examId);
        GetExamMapsReq examsReq = new GetExamMapsReq();
        examsReq.setExamIds(examIds);
        GetExamMapsResp examsResp = examCloudService.getExamMaps(examsReq);
        Map<Long, ExamBean> examMaps = examsResp.getExamMaps();
        ExamBean exam = examMaps.get(examId);
        if (exam != null) {
            info.setExamName(exam.getName());
        }

        return info;
    }

    @Override
    public List<ExamStatisticInfo> overviewForCourse(Long examId, Long courseId) {
        StringBuffer sql = new StringBuffer();
        sql.append(" select");
        sql.append(" course_id,");
        sql.append(" sum(all_count) as allCount,");
        sql.append(" sum(finish_count) as finishCount,");
        sql.append(" sum(pass_score_count) as passScoreCount,");
        sql.append(" sum(good_score_count) as goodScoreCount,");
        sql.append(" sum(illegality_count) as illegalityCount");
        sql.append(" from ec_oe_exam_statistic");
        sql.append(" where exam_id = ").append(examId);
        if (courseId != null) {
            sql.append(" and course_id = ").append(courseId);
        }
        sql.append(" group by course_id");

        Set<Long> courseIds = new HashSet<>();
        List<ExamStatisticInfo> list = jdbcTemplate.query(sql.toString(), new BeanPropertyRowMapper(ExamStatisticInfo.class));
        for (ExamStatisticInfo info : list) {
            courseIds.add(info.getCourseId());

            info.setUnFinishCount(info.getAllCount() - info.getFinishCount());
            if (info.getAllCount() > 0) {
                info.setUnFinishRate(MathUtils.percentage(info.getUnFinishCount(), info.getAllCount()));
                info.setPassScoreRate(MathUtils.percentage(info.getPassScoreCount(), info.getFinishCount()));
                info.setGoodScoreRate(MathUtils.percentage(info.getGoodScoreCount(), info.getFinishCount()));
                info.setIllegalityRate(MathUtils.percentage(info.getIllegalityCount(), info.getFinishCount()));
            }
        }

        // 获取课程信息列表
        GetCoursesByIdListReq req = new GetCoursesByIdListReq();
        req.setCourseIdList(new ArrayList<>(courseIds));
        GetCoursesByIdListResp resp = courseCloudService.getCoursesByIdList(req);
        Map<Long, CourseBean> courseMaps = resp.getCourseList().stream().collect(Collectors.toMap(CourseBean::getId, v -> v));

        for (ExamStatisticInfo info : list) {
            CourseBean course = courseMaps.get(info.getCourseId());
            if (course != null) {
                info.setCourseName(course.getName());
                info.setCourseCode(course.getCode());
            }
        }

        return list;
    }

    @Override
    public List<ExamStatisticInfo> overviewForOrg(Long examId, Long orgId) {
        StringBuffer sql = new StringBuffer();
        sql.append(" select");
        sql.append(" org_id,");
        sql.append(" sum(all_count) as allCount,");
        sql.append(" sum(finish_count) as finishCount,");
        sql.append(" sum(pass_score_count) as passScoreCount,");
        sql.append(" sum(good_score_count) as goodScoreCount,");
        sql.append(" sum(illegality_count) as illegalityCount");
        sql.append(" from ec_oe_exam_statistic");
        sql.append(" where exam_id = ").append(examId);
        if (orgId != null) {
            sql.append(" and org_id = ").append(orgId);
        }
        sql.append(" group by org_id");

        Set<Long> orgIds = new HashSet<>();
        List<ExamStatisticInfo> list = jdbcTemplate.query(sql.toString(), new BeanPropertyRowMapper(ExamStatisticInfo.class));
        for (ExamStatisticInfo info : list) {
            orgIds.add(info.getOrgId());

            info.setUnFinishCount(info.getAllCount() - info.getFinishCount());
            if (info.getAllCount() > 0) {
                info.setUnFinishRate(MathUtils.percentage(info.getUnFinishCount(), info.getAllCount()));
                info.setPassScoreRate(MathUtils.percentage(info.getPassScoreCount(), info.getFinishCount()));
                info.setGoodScoreRate(MathUtils.percentage(info.getGoodScoreCount(), info.getFinishCount()));
                info.setIllegalityRate(MathUtils.percentage(info.getIllegalityCount(), info.getFinishCount()));
            }
        }

        // 获取机构信息列表
        GetOrgMapsReq req = new GetOrgMapsReq();
        req.setOrgIds(orgIds);
        Map<Long, OrgBean> orgMaps = orgCloudService.getOrgMaps(req).getOrgMaps();
        for (ExamStatisticInfo info : list) {
            OrgBean org = orgMaps.get(info.getOrgId());
            if (org != null) {
                info.setOrgName(org.getName());
            }
        }

        return list;
    }

}
